@@ -27,13 +27,10 @@
 
 #include <cerrno>
 #include <cstdlib>
-#include "common/noreturn.h"
 
 using namespace std;
 
-XAPIAN_NORETURN(static void
-parameter_error(const char * param, const string & scheme));
-
+[[noreturn]]
 static void
 parameter_error(const char * msg, const string & scheme)
 {
@@ -68,7 +65,7 @@ set_expansion_scheme(Xapian::Enquire & enq, const map<string, string> & opt)
     if (startswith(scheme, "prob") || startswith(scheme, "trad")) {
 	const char *p = scheme.c_str() + 4;
 	if (*p == '\0') {
-	    enq.set_expansion_scheme("trad");
+	    enq.set_expansion_scheme("prob");
 	    return;
 	}
 	if (C_isspace(*p)) {
@@ -78,7 +75,7 @@ set_expansion_scheme(Xapian::Enquire & enq, const map<string, string> & opt)
 		parameter_error("Parameter k is invalid", scheme);
 	    if (*p)
 		parameter_error("Extra data after first parameter", scheme);
-	    enq.set_expansion_scheme("trad", k);
+	    enq.set_expansion_scheme("prob", k);
 	    return;
 	}
     }
