@@ -4,9 +4,11 @@
 /* Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2001,2005 James Aylett
  * Copyright 2001,2002 Ananova Ltd
- * Copyright 2002-2022 Olly Betts
+ * Copyright 2002-2025 Olly Betts
  * Copyright 2009 Frank J Bruzzaniti
  * Copyright 2012 Mihai Bivol
+ * Copyright 2019 Bruno Baruffaldi
+ * Copyright 2020 Parth Kapadia
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -37,6 +39,7 @@
 
 #include <sys/types.h>
 #include "safeunistd.h"
+#include <cassert>
 #include <cerrno>
 #include <cstdio>
 #include <cstdlib>
@@ -46,28 +49,31 @@
 
 #include <xapian.h>
 
+#include "abiwordparser.h"
 #include "append_filename_arg.h"
-#include "atomparse.h"
+#include "atomparser.h"
+#include "datetime.h"
 #include "diritor.h"
 #include "failed.h"
+#include "gnumericparser.h"
+#include "hashterm.h"
+#include "htmlparser.h"
 #include "md5wrap.h"
-#include "metaxmlparse.h"
-#include "msxmlparse.h"
-#include "myhtmlparse.h"
-#include "opendocparse.h"
+#include "msxmlparser.h"
+#include "opendocmetaparser.h"
+#include "opendocparser.h"
 #include "pkglibbindir.h"
 #include "runfilter.h"
 #include "sample.h"
 #include "str.h"
 #include "stringutils.h"
-#include "svgparse.h"
+#include "svgparser.h"
 #include "tmpdir.h"
 #include "utf8convert.h"
-#include "utils.h"
 #include "values.h"
-#include "xmlparse.h"
-#include "xlsxparse.h"
-#include "xpsxmlparse.h"
+#include "worker.h"
+#include "xlsxparser.h"
+#include "xpsparser.h"
 
 using namespace std;
 
@@ -123,6 +129,22 @@ skip(const string& urlterm, const string& context, const string& msg,
     cout << "Skipping - " << msg << endl;
 }
 
+static void
+skip_cmd_failed(const string& urlterm, const string& context,
+		const char* const cmd[],
+		off_t size, time_t last_mod)
+{
+    string message;
+    const char* sep = "['";
+    for (auto i = cmd; *i; ++i) {
+	message += sep;
+	message += *i;
+	sep = "', '";
+    }
+    message += "'] failed";
+    skip(urlterm, context, message, size, last_mod);
+}
+
 static void
 skip_cmd_failed(const string& urlterm, const string& context, const string& cmd,
 		off_t size, time_t last_mod)
@@ -145,26 +167,187 @@ skip_unknown_mimetype(const string& urlterm, const string& context,
 	 size, last_mod);
 }
 
+void
+index_add_default_libraries()
+{
+#if defined HAVE_POPPLER
+    Worker* omindex_poppler = new Worker("omindex_poppler");
+    index_library("application/pdf", omindex_poppler);
+#endif
+#if defined HAVE_LIBEBOOK
+    Worker* omindex_libebook = new Worker("omindex_libebook");
+    index_library("application/vnd.palm", omindex_libebook);
+    index_library("application/x-fictionbook+xml", omindex_libebook);
+    index_library("application/x-zip-compressed-fb2", omindex_libebook);
+    index_library("application/x-sony-bbeb", omindex_libebook);
+    index_library("application/x-tcr-ebook", omindex_libebook);
+    index_library("application/x-qioo-ebook", omindex_libebook);
+#endif
+#if defined HAVE_LIBETONYEK
+    Worker* omindex_libetonyek = new Worker("omindex_libetonyek");
+    index_library("application/vnd.apple.keynote", omindex_libetonyek);
+    index_library("application/vnd.apple.pages", omindex_libetonyek);
+    index_library("application/vnd.apple.numbers", omindex_libetonyek);
+#endif
+#if defined HAVE_LIBGEPUB
+    Worker* omindex_libgepub = new Worker("omindex_libgepub");
+    index_library("application/epub+zip", omindex_libgepub);
+#endif
+#if defined HAVE_TESSERACT
+    Worker* omindex_tesseract = new Worker("omindex_tesseract");
+    index_library("image/gif", omindex_tesseract);
+    index_library("image/jpeg", omindex_tesseract);
+    index_library("image/png", omindex_tesseract);
+    index_library("image/webp", omindex_tesseract);
+    index_library("image/tiff", omindex_tesseract);
+    index_library("image/x-portable-bitmap", omindex_tesseract);
+    index_library("image/x-portable-graymap", omindex_tesseract);
+    index_library("image/x-portable-anymap", omindex_tesseract);
+    index_library("image/x-portable-pixmap", omindex_tesseract);
+#endif
+#if defined HAVE_GMIME
+    Worker* omindex_gmime = new Worker("omindex_gmime");
+    index_library("message/rfc822", omindex_gmime);
+    index_library("message/news", omindex_gmime);
+#endif
+#if defined HAVE_LIBARCHIVE
+    Worker* omindex_libarchive = new Worker("omindex_libarchive");
+    index_library("application/oxps", omindex_libarchive);
+    index_library("application/vnd.ms-xpsdocument", omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.text",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.spreadsheet",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.presentation",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.graphics",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.chart",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.formula",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.database",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.image",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.text-master",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.text-template",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.spreadsheet-template",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.presentation-template",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.graphics-template",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.chart-template",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.formula-template",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.image-template",
+		  omindex_libarchive);
+    index_library("application/vnd.oasis.opendocument.text-web",
+		  omindex_libarchive);
+    index_library("application/vnd.sun.xml.calc",
+		  omindex_libarchive);
+    index_library("application/vnd.sun.xml.calc.template",
+		  omindex_libarchive);
+    index_library("application/vnd.sun.xml.draw",
+		  omindex_libarchive);
+    index_library("application/vnd.sun.xml.draw.template",
+		  omindex_libarchive);
+    index_library("application/vnd.sun.xml.impress",
+		  omindex_libarchive);
+    index_library("application/vnd.sun.xml.impress.template",
+		  omindex_libarchive);
+    index_library("application/vnd.sun.xml.math",
+		  omindex_libarchive);
+    index_library("application/vnd.sun.xml.writer",
+		  omindex_libarchive);
+    index_library("application/vnd.sun.xml.writer.global",
+		  omindex_libarchive);
+    index_library("application/vnd.sun.xml.writer.template",
+		  omindex_libarchive);
+    index_library("application/vnd.openxmlformats-officedocument."
+		  "wordprocessingml.document", omindex_libarchive);
+    index_library("application/vnd.openxmlformats-officedocument."
+		  "wordprocessingml.template", omindex_libarchive);
+    index_library("application/vnd.openxmlformats-officedocument."
+		  "spreadsheetml.sheet", omindex_libarchive);
+    index_library("application/vnd.openxmlformats-officedocument."
+		  "spreadsheetml.template", omindex_libarchive);
+    index_library("application/vnd.openxmlformats-officedocument."
+		  "presentationml.presentation", omindex_libarchive);
+    index_library("application/vnd.openxmlformats-officedocument."
+		  "presentationml.slideshow", omindex_libarchive);
+    index_library("application/vnd.openxmlformats-officedocument."
+		  "presentationml.template", omindex_libarchive);
+#endif
+#if defined HAVE_LIBABW
+    Worker* omindex_libabw = new Worker("omindex_libabw");
+    index_library("application/x-abiword", omindex_libabw);
+    index_library("application/x-abiword-compressed", omindex_libabw);
+#endif
+#if defined HAVE_LIBCDR
+    Worker* omindex_libcdr = new Worker("omindex_libcdr");
+    index_library("image/x-coreldraw", omindex_libcdr);
+#endif
+#if defined HAVE_LIBEXTRACTOR
+    Worker* omindex_libextractor = new Worker("omindex_libextractor");
+    index_library("video/mpeg", omindex_libextractor);
+    index_library("video/x-flv", omindex_libextractor);
+    index_library("video/x-msvideo", omindex_libextractor);
+    index_library("video/x-ms-asf", omindex_libextractor);
+    index_library("video/quicktime", omindex_libextractor);
+    index_library("video/ogg", omindex_libextractor);
+    index_library("audio/flac", omindex_libextractor);
+    index_library("audio/mpeg", omindex_libextractor);
+    index_library("audio/ogg", omindex_libextractor);
+    index_library("audio/x-wav", omindex_libextractor);
+    index_library("audio/x-mod", omindex_libextractor);
+    index_library("audio/x-s3m", omindex_libextractor);
+#endif
+#if defined HAVE_LIBMWAW
+    Worker* omindex_libmwaw = new Worker("omindex_libmwaw");
+    index_library("application/clarisworks", omindex_libmwaw);
+    index_library("image/x-pict", omindex_libmwaw);
+#endif
+}
+
 void
 index_add_default_filters()
 {
-    index_command("application/msword", Filter("antiword -mUTF-8.txt", false));
+    // Command needs to be run using /bin/sh.
+    auto USE_SHELL = Filter::USE_SHELL;
+    // Currently none of these commands needs USE_SHELL.
+    (void)USE_SHELL;
+    // Input should be piped to stdin.
+    auto PIPE_IN = Filter::PIPE_IN;
+    // Filename can be /dev/stdin (which must be seekable).
+    auto SEEK_DEV_STDIN = Filter::SEEK_DEV_STDIN;
+    // Filename can be /dev/stdin (which can be a pipe).
+    auto PIPE_DEV_STDIN = Filter::PIPE_DEV_STDIN;
+    index_command("application/msword",
+		  Filter("antiword -mUTF-8.txt -", PIPE_IN));
     index_command("application/vnd.ms-excel",
-		  Filter("xls2csv -c' ' -q0 -dutf-8", false));
+		  Filter("xls2csv -c' ' -q0 -dutf-8", PIPE_DEV_STDIN));
     index_command("application/vnd.ms-powerpoint",
-		  Filter("catppt -dutf-8", false));
+		  Filter("catppt -dutf-8", PIPE_DEV_STDIN));
     // Looking at the source of wpd2html and wpd2text I think both output
     // UTF-8, but it's hard to be sure without sample Unicode .wpd files
     // as they don't seem to be at all well documented.
-    index_command("application/vnd.wordperfect", Filter("wpd2text", false));
+    index_command("application/vnd.wordperfect",
+		  Filter("wpd2text", SEEK_DEV_STDIN));
     // wps2text produces UTF-8 output from the sample files I've tested.
-    index_command("application/vnd.ms-works", Filter("wps2text", false));
+    index_command("application/vnd.ms-works",
+		  Filter("wps2text", SEEK_DEV_STDIN));
     // Output is UTF-8 according to "man djvutxt".  Generally this seems to
     // be true, though some examples from djvu.org generate isolated byte
     // 0x95 in a context which suggests it might be intended to be a bullet
     // (as it is in CP1252).
-    index_command("image/vnd.djvu", Filter("djvutxt", false));
-    index_command("text/markdown", Filter("markdown", "text/html", false));
+    index_command("image/vnd.djvu", Filter("djvutxt -", PIPE_IN));
+    index_command("text/markdown",
+		  Filter("markdown", "text/html", PIPE_IN));
     // The --text option unhelpfully converts all non-ASCII characters to "?"
     // so we use --html instead, which produces HTML entities.  The --nopict
     // option suppresses exporting picture files as pictNNNN.wmf in the current
@@ -172,59 +355,60 @@ index_add_default_filters()
     // but it was fixed in unrtf 0.20.4.
     index_command("application/rtf",
 		  Filter("unrtf --nopict --html 2>/dev/null", "text/html",
-			 false));
+			 PIPE_IN));
     index_command("text/rtf",
 		  Filter("unrtf --nopict --html 2>/dev/null", "text/html",
-			 false));
-    index_command("text/x-rst", Filter("rst2html", "text/html", false));
+			 PIPE_IN));
+    index_command("text/x-rst",
+		  Filter("rst2html", "text/html", PIPE_IN));
     index_command("application/x-mspublisher",
-		  Filter("pub2xhtml", "text/html", false));
+		  Filter("pub2xhtml", "text/html", SEEK_DEV_STDIN));
     index_command("application/vnd.ms-outlook",
-		  Filter(get_pkglibbindir() + "/outlookmsg2html", "text/html",
-			 false));
+		  Filter(get_pkglibbindir() + "/outlookmsg2html",
+			 "text/html", SEEK_DEV_STDIN));
     index_command("application/vnd.ms-visio.drawing",
-		  Filter("vsd2xhtml", "image/svg+xml", false));
+		  Filter("vsd2xhtml", "image/svg+xml", SEEK_DEV_STDIN));
     index_command("application/vnd.ms-visio.stencil",
-		  Filter("vsd2xhtml", "image/svg+xml", false));
+		  Filter("vsd2xhtml", "image/svg+xml", SEEK_DEV_STDIN));
     index_command("application/vnd.ms-visio.template",
-		  Filter("vsd2xhtml", "image/svg+xml", false));
+		  Filter("vsd2xhtml", "image/svg+xml", SEEK_DEV_STDIN));
     index_command("application/vnd.visio",
-		  Filter("vsd2xhtml", "image/svg+xml", false));
+		  Filter("vsd2xhtml", "image/svg+xml", SEEK_DEV_STDIN));
     // pod2text's output character set doesn't seem to be documented, but from
     // inspecting the source it looks like it's probably iso-8859-1.  We need
     // to pass "--errors=stderr" or else minor POD formatting errors cause a
     // file not to be indexed.
     index_command("text/x-perl",
 		  Filter("pod2text --errors=stderr",
-			 "text/plain", "iso-8859-1", false));
+			 "text/plain", "iso-8859-1", PIPE_IN));
     // FIXME: -e0 means "UTF-8", but that results in "fi", "ff", "ffi", etc
     // appearing as single ligatures.  For European languages, it's actually
     // better to use -e2 (ISO-8859-1) and then convert, so let's do that for
     // now until we handle Unicode "compatibility decompositions".
     index_command("application/x-dvi",
-		  Filter("catdvi -e2 -s", "text/plain", "iso-8859-1", false));
+		  Filter("catdvi -e2 -s", "text/plain", "iso-8859-1", PIPE_IN));
     // Simplistic - ought to look in index.rdf files for filename and character
     // set.
     index_command("application/x-maff",
 		  Filter("unzip -p %f '*/*.*htm*'", "text/html", "iso-8859-1",
-			 false));
+			 SEEK_DEV_STDIN));
     index_command("application/x-mimearchive",
 		  Filter(get_pkglibbindir() + "/mhtml2html", "text/html",
-			 false));
+			 PIPE_DEV_STDIN));
     index_command("message/news",
 		  Filter(get_pkglibbindir() + "/rfc822tohtml", "text/html",
-			 false));
+			 PIPE_DEV_STDIN));
     index_command("message/rfc822",
 		  Filter(get_pkglibbindir() + "/rfc822tohtml", "text/html",
-			 false));
+			 PIPE_DEV_STDIN));
     index_command("text/vcard",
-		  Filter(get_pkglibbindir() + "/vcard2text", false));
+		  Filter(get_pkglibbindir() + "/vcard2text", PIPE_DEV_STDIN));
     index_command("application/vnd.apple.keynote",
-		  Filter("key2text", false));
+		  Filter("key2text", SEEK_DEV_STDIN));
     index_command("application/vnd.apple.numbers",
-		  Filter("numbers2text", false));
+		  Filter("numbers2text", SEEK_DEV_STDIN));
     index_command("application/vnd.apple.pages",
-		  Filter("pages2text", false));
+		  Filter("pages2text", SEEK_DEV_STDIN));
 }
 
 void
@@ -360,7 +544,10 @@ get_pdf_metainfo(int fd, string& author, string& title,
 {
     try {
 	string pdfinfo;
-	run_filter(fd, "pdfinfo -enc UTF-8 -", false, &pdfinfo);
+	static const char* const cmd[] = {
+	    "pdfinfo", "-enc", "UTF-8", "-", NULL
+	};
+	run_filter(fd, cmd);
 	parse_pdf_metainfo(pdfinfo, author, title, keywords, topic, pages);
     } catch (const ReadError&) {
 	// It's probably best to index the document even if pdfinfo fails.
@@ -372,9 +559,11 @@ get_pdf_metainfo(const string& file, string& author, string& title,
 		 string& keywords, string& topic, int& pages)
 {
     try {
-	string cmd = "pdfinfo -enc UTF-8";
-	append_filename_argument(cmd, file);
-	parse_pdf_metainfo(stdout_to_string(cmd, false),
+	const char* cmd[] = {
+	    "pdfinfo", "-enc", "UTF-8", NULL, NULL
+	};
+	cmd[3] = file.c_str();
+	parse_pdf_metainfo(stdout_to_string(cmd),
 			   author, title, keywords, topic, pages);
     } catch (const ReadError&) {
 	// It's probably best to index the document even if pdfinfo fails.
@@ -531,9 +720,9 @@ index_add_document(const string& urlterm, time_t last_altered,
 void
 index_mimetype(const string& file, const string& urlterm, const string& url,
 	       const string& ext,
-	       const string& mimetype,
+	       string mimetype,
 	       DirectoryIterator& d,
-	       Xapian::Document & newdocument,
+	       string pathterm,
 	       string record)
 {
     string context(file, root.size(), string::npos);
@@ -566,9 +755,39 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 	}
     }
 
-    if (verbose) cout << flush;
+    // If we didn't get the mime type from the extension, call libmagic to get
+    // it.
+    if (mimetype.empty()) {
+	mimetype = d.get_magic_mimetype();
+	if (mimetype.empty()) {
+	    skip(urlterm, file.substr(root.size()),
+		 "Unknown extension and unrecognised format",
+		 d.get_size(), d.get_mtime(), SKIP_SHOW_FILENAME);
+	    return;
+	}
+    }
+
+    if (verbose)
+	cout << "Indexing \"" << file.substr(root.size()) << "\" as "
+	     << mimetype << " ... " << flush;
+
+    // Use `file` as the basis, as we don't want URL encoding in these terms,
+    // but need to switch over the initial part so we get `/~olly/foo/bar` not
+    // `/home/olly/public_html/foo/bar`.
+    Xapian::Document newdocument;
+    size_t j;
+    while ((j = pathterm.rfind('/')) > 1 && j != string::npos) {
+	pathterm.resize(j);
+	if (pathterm.length() > MAX_SAFE_TERM_LENGTH) {
+	    string term_hash = hash_long_term(pathterm, MAX_SAFE_TERM_LENGTH);
+	    newdocument.add_boolean_term(term_hash);
+	} else {
+	    newdocument.add_boolean_term(pathterm);
+	}
+    }
 
     string author, title, sample, keywords, topic, dump;
+    string to, cc, bcc, message_id;
     string md5;
     time_t created = time_t(-1);
     int pages = -1;
@@ -589,7 +808,28 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 	}
     }
     try {
-	if (cmd_it != commands.end()) {
+	if (cmd_it != commands.end() && cmd_it->second.worker) {
+	    // Use a worker process to extract the content.
+	    Worker* wrk = cmd_it->second.worker;
+	    int r = wrk->extract(file, mimetype, dump, title, keywords, author,
+				 to, cc, bcc, message_id, pages, created);
+	    if (r != 0) {
+		string msg = wrk->get_error();
+		assert(!msg.empty());
+		skip(urlterm, context, msg, d.get_size(), d.get_mtime());
+		if (r < 0) {
+		    // Hard failure - don't try this filter again for this run.
+		    string filter_entry;
+		    if (cmd_it != commands.end()) {
+			filter_entry = cmd_it->first;
+		    } else {
+			filter_entry = mimetype;
+		    }
+		    commands[filter_entry] = Filter();
+		}
+		return;
+	    }
+	} else if (cmd_it != commands.end()) {
 	    // Easy "run a command and read text or HTML from stdout or a
 	    // temporary file" cases.
 	    auto& filter = cmd_it->second;
@@ -610,6 +850,7 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 		return;
 	    }
 	    bool use_shell = filter.use_shell();
+	    bool input_on_stdin = filter.input_on_stdin();
 	    bool substituted = false;
 	    string tmpout;
 	    size_t pcent = 0;
@@ -623,6 +864,11 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 			break;
 		    case 'f': { // %f -> escaped filename.
 			substituted = true;
+			if (filter.dev_stdin()) {
+			    cmd.replace(pcent, 2, "/dev/stdin",
+					CONST_STRLEN("/dev/stdin"));
+			    break;
+			}
 			string tail(cmd, pcent + 2);
 			cmd.resize(pcent);
 			// Suppress the space append_filename_argument()
@@ -666,13 +912,23 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 		}
 	    }
 	    if (!substituted && cmd != "true") {
-		// If no %f, append the filename to the command.
-		append_filename_argument(cmd, file);
+		if (input_on_stdin) {
+		    if (filter.dev_stdin()) {
+			cmd += " /dev/stdin";
+		    }
+		} else {
+		    // If no %f, append the filename to the command.
+		    append_filename_argument(cmd, file);
+		}
 	    }
 	    try {
 		if (!tmpout.empty()) {
 		    // Output in temporary file.
-		    run_filter(cmd, use_shell);
+		    if (input_on_stdin) {
+			run_filter(d.get_fd(), cmd, use_shell);
+		    } else {
+			run_filter(cmd, use_shell);
+		    }
 		    if (!load_file(tmpout, dump, NOCACHE)) {
 			throw ReadError("Couldn't read output file");
 		    }
@@ -682,20 +938,24 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 		    // filing system.
 		} else {
 		    // Output on stdout.
-		    run_filter(cmd, use_shell, &dump);
+		    if (input_on_stdin) {
+			run_filter(d.get_fd(), cmd, use_shell, &dump);
+		    } else {
+			run_filter(cmd, use_shell, &dump);
+		    }
 		}
 		const string& charset = filter.output_charset;
 		if (filter.output_type == "text/html") {
-		    MyHtmlParser p;
+		    HtmlParser p;
 		    p.ignore_metarobots();
 		    p.description_as_sample = description_as_sample;
 		    try {
-			p.parse_html(dump, charset, false);
+			p.parse(dump, charset, false);
 		    } catch (const string& newcharset) {
 			p.reset();
 			p.ignore_metarobots();
 			p.description_as_sample = description_as_sample;
-			p.parse_html(dump, newcharset, true);
+			p.parse(dump, newcharset, true);
 		    } catch (const ReadError&) {
 			skip_cmd_failed(urlterm, context, cmd,
 					d.get_size(), d.get_mtime());
@@ -726,18 +986,18 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 	    }
 	} else if (mimetype == "text/html" || mimetype == "text/x-php") {
 	    const string& text = d.file_to_string();
-	    MyHtmlParser p;
+	    HtmlParser p;
 	    if (ignore_exclusions) p.ignore_metarobots();
 	    p.description_as_sample = description_as_sample;
 	    try {
 		// Default HTML character set is latin 1, though not specifying
 		// one is deprecated these days.
-		p.parse_html(text, "iso-8859-1", false);
+		p.parse(text, "iso-8859-1", false);
 	    } catch (const string& newcharset) {
 		p.reset();
 		if (ignore_exclusions) p.ignore_metarobots();
 		p.description_as_sample = description_as_sample;
-		p.parse_html(text, newcharset, true);
+		p.parse(text, newcharset, true);
 	    }
 	    if (!p.indexing_allowed) {
 		skip_meta_tag(urlterm, context,
@@ -772,9 +1032,11 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 		// FIXME: What charset is the file?  Look at contents?
 	    }
 	} else if (mimetype == "application/pdf") {
-	    const char* cmd = "pdftotext -enc UTF-8 - -";
+	    const char* const cmd[] = {
+		"pdftotext", "-enc", "UTF-8", "-", "-", NULL
+	    };
 	    try {
-		run_filter(d.get_fd(), cmd, false, &dump);
+		run_filter(d.get_fd(), cmd, &dump);
 	    } catch (const ReadError&) {
 		skip_cmd_failed(urlterm, context, cmd,
 				d.get_size(), d.get_mtime());
@@ -799,14 +1061,17 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 		     d.get_size(), d.get_mtime());
 		return;
 	    }
-	    string cmd = "ps2pdf -";
-	    append_filename_argument(cmd, tmpfile);
+	    const char* cmd[] = {
+		"ps2pdf", "-", NULL, NULL
+	    };
+	    cmd[2] = tmpfile.c_str();
 	    try {
-		run_filter(d.get_fd(), cmd, false);
-		cmd = "pdftotext -enc UTF-8";
-		append_filename_argument(cmd, tmpfile);
-		cmd += " -";
-		run_filter(cmd, false, &dump);
+		run_filter(d.get_fd(), cmd);
+		const char* cmd2[] = {
+		    "pdftotext", "-enc", "UTF-8", NULL, "-", NULL
+		};
+		cmd2[3] = tmpfile.c_str();
+		run_filter(cmd2, &dump);
 	    } catch (const ReadError&) {
 		skip_cmd_failed(urlterm, context, cmd,
 				d.get_size(), d.get_mtime());
@@ -843,18 +1108,19 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 		return;
 	    }
 
-	    cmd = "unzip -p";
-	    append_filename_argument(cmd, file);
-	    cmd += " meta.xml";
+	    const char* cmd2[] = {
+		"unzip", "-p", NULL, "meta.xml", NULL
+	    };
+	    cmd2[2] = file.c_str();
 	    try {
-		MetaXmlParser metaxmlparser;
-		metaxmlparser.parse(stdout_to_string(cmd, false));
-		title = metaxmlparser.title;
-		keywords = metaxmlparser.keywords;
-		// FIXME: topic = metaxmlparser.topic;
-		sample = metaxmlparser.sample;
-		author = metaxmlparser.author;
-		pages = metaxmlparser.pages;
+		OpenDocMetaParser metaparser;
+		metaparser.parse(stdout_to_string(cmd2));
+		title = metaparser.title;
+		keywords = metaparser.keywords;
+		// FIXME: topic = metaparser.topic;
+		sample = metaparser.sample;
+		author = metaparser.author;
+		pages = metaparser.pages;
 	    } catch (const ReadError&) {
 		// It's probably best to index the document even if this fails.
 	    }
@@ -864,9 +1130,7 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 	    const char* args = NULL;
 	    string tail(mimetype, 46);
 	    if (startswith(tail, "wordprocessingml.")) {
-		// unzip returns exit code 11 if a file to extract wasn't found
-		// which we want to ignore, because there may be no headers or
-		// no footers.
+		// There may be no headers or no footers.
 		args = " word/document.xml"
 		       " 'word/header*.xml'"
 		       " 'word/footer*.xml'"
@@ -891,9 +1155,7 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 		    return;
 		}
 	    } else if (startswith(tail, "presentationml.")) {
-		// unzip returns exit code 11 if a file to extract wasn't found
-		// which we want to ignore, because there may be no notesSlides
-		// or comments.
+		// There may be no notesSlides or no comments.
 		args = " 'ppt/slides/slide*.xml'"
 		       " 'ppt/notesSlides/notesSlide*.xml'"
 		       " 'ppt/comments/comment*.xml'"
@@ -914,7 +1176,7 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 		    // Treat exit status 11 from unzip as success - this is
 		    // what we get if one of the listed filenames to extract
 		    // doesn't match anything in the zip file.
-		    xmlparser.parse_xml(stdout_to_string(cmd, false, 11));
+		    xmlparser.parse(stdout_to_string(cmd, false, 11));
 		    dump = xmlparser.dump;
 		} catch (const ReadError&) {
 		    skip_cmd_failed(urlterm, context, cmd,
@@ -923,48 +1185,59 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 		}
 	    }
 
-	    string cmd = "unzip -p";
-	    append_filename_argument(cmd, file);
-	    cmd += " docProps/core.xml";
+	    const char* cmd[] = {
+		"unzip", "-p", NULL, "docProps/core.xml", NULL
+	    };
+	    cmd[2] = file.c_str();
 	    try {
-		MetaXmlParser metaxmlparser;
-		metaxmlparser.parse(stdout_to_string(cmd, false));
-		title = metaxmlparser.title;
-		keywords = metaxmlparser.keywords;
-		// FIXME: topic = metaxmlparser.topic;
-		sample = metaxmlparser.sample;
-		author = metaxmlparser.author;
+		OpenDocMetaParser metaparser;
+		metaparser.parse(stdout_to_string(cmd));
+		title = metaparser.title;
+		keywords = metaparser.keywords;
+		// FIXME: topic = metaparser.topic;
+		sample = metaparser.sample;
+		author = metaparser.author;
 	    } catch (const ReadError&) {
 		// It's probably best to index the document even if this fails.
 	    }
 	} else if (mimetype == "application/x-abiword") {
-	    XmlParser xmlparser;
-	    const string & text = d.file_to_string();
-	    xmlparser.parse_xml(text);
-	    dump = xmlparser.dump;
-	    title = xmlparser.title;
-	    keywords = xmlparser.keywords;
-	    sample = xmlparser.sample;
-	    author = xmlparser.author;
-	    created = xmlparser.created;
+	    AbiwordParser abiwordparser;
+	    const string& text = d.file_to_string();
+	    abiwordparser.parse(text);
+	    dump = abiwordparser.dump;
+	    title = abiwordparser.title;
+	    keywords = abiwordparser.keywords;
+	    sample = abiwordparser.sample;
+	    author = abiwordparser.author;
+	    created = abiwordparser.created;
 	    md5_string(text, md5);
 	} else if (mimetype == "application/x-abiword-compressed") {
-	    XmlParser xmlparser;
-	    xmlparser.parse_xml(d.gzfile_to_string());
-	    dump = xmlparser.dump;
-	    title = xmlparser.title;
-	    keywords = xmlparser.keywords;
-	    sample = xmlparser.sample;
-	    author = xmlparser.author;
-	    created = xmlparser.created;
+	    AbiwordParser abiwordparser;
+	    abiwordparser.parse(d.gzfile_to_string());
+	    dump = abiwordparser.dump;
+	    title = abiwordparser.title;
+	    keywords = abiwordparser.keywords;
+	    sample = abiwordparser.sample;
+	    author = abiwordparser.author;
+	    created = abiwordparser.created;
+	} else if (mimetype == "application/x-gnumeric") {
+	    GnumericParser p;
+	    p.parse(d.gzfile_to_string());
+	    dump = p.dump;
+	    title = p.title;
+	    keywords = p.keywords;
+	    sample = p.sample;
+	    author = p.author;
+	    created = p.created;
 	} else if (mimetype == "application/oxps" ||
 		   mimetype == "application/vnd.ms-xpsdocument") {
-	    string cmd = "unzip -p";
-	    append_filename_argument(cmd, file);
-	    cmd += " 'Documents/*/Pages/*.fpage'";
+	    const char* cmd[] = {
+		"unzip", "-p", NULL, "Documents/*/Pages/*.fpage", NULL
+	    };
+	    cmd[2] = file.c_str();
 	    try {
-		XpsXmlParser xpsparser;
-		run_filter(cmd, false, &dump);
+		XpsParser xpsparser;
+		run_filter(cmd,  &dump);
 		xpsparser.parse(dump);
 		dump = xpsparser.dump;
 	    } catch (const ReadError&) {
@@ -973,12 +1246,13 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 		return;
 	    }
 
-	    cmd = "unzip -p";
-	    append_filename_argument(cmd, file);
-	    cmd += " docProps/core.xml";
+	    const char* cmd2[] = {
+		"unzip", "-p", NULL, "docProps/core.xml", NULL
+	    };
+	    cmd2[2] = file.c_str();
 	    try {
-		MetaXmlParser metaparser;
-		metaparser.parse(stdout_to_string(cmd, false));
+		OpenDocMetaParser metaparser;
+		metaparser.parse(stdout_to_string(cmd2));
 		title = metaparser.title;
 		keywords = metaparser.keywords;
 		// FIXME: topic = metaparser.topic;
@@ -1029,9 +1303,11 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 	    author = svgparser.author;
 	} else if (mimetype == "application/vnd.debian.binary-package" ||
 		   mimetype == "application/x-debian-package") {
-	    const char* cmd = "dpkg-deb -f - Description";
+	    const char* const cmd[] = {
+		"dpkg-deb", "-f", "-", "Description", NULL
+	    };
 	    string desc;
-	    run_filter(d.get_fd(), cmd, false, &desc);
+	    run_filter(d.get_fd(), cmd, &desc);
 	    // First line is short description, which we use as the title.
 	    string::size_type idx = desc.find('\n');
 	    title.assign(desc, 0, idx);
@@ -1040,10 +1316,13 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 	    }
 	} else if (mimetype == "application/x-redhat-package-manager" ||
 		   mimetype == "application/x-rpm") {
-	    string cmd("rpm -q --qf '%{SUMMARY}\\n%{DESCRIPTION}' -p");
-	    append_filename_argument(cmd, file);
+	    const char* cmd[] = {
+		"rpm", "-q", "--qf", "%{SUMMARY}\\n%{DESCRIPTION}", "-p",
+		NULL, NULL
+	    };
+	    cmd[5] = file.c_str();
 	    string desc;
-	    run_filter(cmd, false, &desc);
+	    run_filter(cmd, &desc);
 	    // First line is summary, which we use as the title.
 	    string::size_type idx = desc.find('\n');
 	    title.assign(desc, 0, idx);
@@ -1088,7 +1367,7 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 	// If dump contain only formfeeds, then trim_end will be string::npos
 	// and ++trim_end will be 0, which is the correct new size.
 	string::size_type trim_end = dump.find_last_not_of('\f');
-	if (++trim_end != dump.size())
+	if (UNSIGNED_OVERFLOW_OK(++trim_end) != dump.size())
 	    dump.resize(trim_end);
 
 	if (dump.empty()) {
@@ -1131,6 +1410,22 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 	    record += "\nauthor=";
 	    record += author;
 	}
+	if (!to.empty()) {
+	    record += "\nto=";
+	    record += to;
+	}
+	if (!cc.empty()) {
+	    record += "\ncc=";
+	    record += cc;
+	}
+	if (!bcc.empty()) {
+	    record += "\nbcc=";
+	    record += bcc;
+	}
+	if (!message_id.empty()) {
+	    record += "\nmsgid=";
+	    record += message_id;
+	}
 	record += "\ntype=";
 	record += mimetype;
 	time_t mtime = d.get_mtime();
@@ -1195,6 +1490,25 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 	    indexer.index_text(author, 1, "A");
 	}
 
+	if (!to.empty()) {
+	    indexer.increase_termpos(100);
+	    indexer.index_text(to, 1, "XTO");
+	}
+
+	if (!cc.empty()) {
+	    indexer.increase_termpos(100);
+	    indexer.index_text(cc, 1, "XCC");
+	}
+
+	if (!bcc.empty()) {
+	    indexer.increase_termpos(100);
+	    indexer.index_text(bcc, 1, "XBCC");
+	}
+
+	if (!message_id.empty()) {
+	    newdocument.add_boolean_term("XMID:" + message_id);
+	}
+
 	// mimeType:
 	newdocument.add_boolean_term("T" + mimetype);
 
@@ -1238,6 +1552,12 @@ index_mimetype(const string& file, const string& urlterm, const string& url,
 	newdocument.add_value(VALUE_SIZE,
 			      Xapian::sortable_serialise(size));
 
+	if (created != static_cast<time_t>(-1)) {
+	    // Add created time as a value to allow "sort by created date".
+	    newdocument.add_value(VALUE_CREATED,
+				  int_to_binary_string(uint32_t(created)));
+	}
+
 	bool inc_tag_added = false;
 	if (d.is_other_readable()) {
 	    inc_tag_added = true;
