@@ -51,10 +51,13 @@
 #include "hashterm.h"
 #include "index_file.h"
 #include "mime.h"
+#include "parseint.h"
+#include "pkglibbindir.h"
 #include "realtime.h"
 #include "str.h"
 #include "stringutils.h"
 #include "urlencode.h"
+#include "worker.h"
 
 #include "gnu_getopt.h"
 
@@ -71,7 +74,7 @@ static off_t max_size = 0;
 static std::string pretty_max_size;
 static bool verbose = false;
 static double sleep_before_opendir = 0;
-static bool date_terms = true;
+static bool date_terms = false;
 
 static string root;
 static string url_start_path;
@@ -160,41 +163,11 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 	return;
     }
 
-    // If we didn't get the mime type from the extension, call libmagic to get
-    // it.
-    if (mimetype.empty()) {
-	mimetype = d.get_magic_mimetype();
-	if (mimetype.empty()) {
-	    skip(urlterm, file.substr(root.size()), "Unknown extension and unrecognised format",
-		 d.get_size(), d.get_mtime(), SKIP_SHOW_FILENAME);
-	    return;
-	}
-    }
-
-    if (verbose)
-	cout << "Indexing \"" << file.substr(root.size()) << "\" as "
-	     << mimetype << " ... ";
-
-    Xapian::Document new_doc;
-
-    // Use `file` as the basis, as we don't want URL encoding in these terms,
-    // but need to switch over the initial part so we get `/~olly/foo/bar` not
-    // `/home/olly/public_html/foo/bar`.
     string path_term("P");
     path_term += url_start_path;
     path_term.append(file, root.size(), string::npos);
 
-    size_t i;
-    while ((i = path_term.rfind('/')) > 1 && i != string::npos) {
-	path_term.resize(i);
-	if (path_term.length() > MAX_SAFE_TERM_LENGTH) {
-	    new_doc.add_boolean_term(hash_long_term(path_term, MAX_SAFE_TERM_LENGTH));
-	} else {
-	    new_doc.add_boolean_term(path_term);
-	}
-    }
-
-    index_mimetype(file, urlterm, url, ext, mimetype, d, new_doc, string());
+    index_mimetype(file, urlterm, url, ext, mimetype, d, path_term, string());
 }
 
 static void
@@ -336,14 +309,42 @@ parse_filter_rule(const char* rule, map<string, string>& mime_map)
     }
 
     const char* cmd = s + 1;
+    unsigned flags = 0;
+    if (cmd[0] == '|') {
+	flags |= Filter::PIPE_IN;
+	++cmd;
+	// FIXME: Do we need a way to set PIPE_DEV_STDIN and SEEK_DEV_STDIN?
+	//
+	// PIPE_DEV_STDIN doesn't seem to offer much over |foo2txt /dev/stdin
+	// for user-specified filters (internally it provides a way to
+	// gracefully handle platforms without /dev/stdin).
+	//
+	// SEEK_DEV_STDIN isn't currently easily approximated though.
+    }
     // Analyse the command string to decide if it needs a shell.
-    bool use_shell = command_needs_shell(cmd);
+    if (command_needs_shell(cmd))
+	flags |= Filter::USE_SHELL;
     index_command(string(rule, c - rule),
-		  Filter(string(cmd), output_type, output_charset, use_shell));
+		  Filter(string(cmd), output_type, output_charset, flags));
 
     return true;
 }
 
+static bool
+parse_worker_rule(const char* rule)
+{
+    const char* s = strchr(rule, ':');
+    if (s == NULL || s[1] == '\0') {
+	cerr << "Invalid worker mapping '" << rule << "'\n"
+		"Should be of the form TYPE:WORKER\n"
+		"e.g. 'application/msword:omindex_libreofficekit\n";
+	return false;
+    }
+
+    index_library(string(rule, s - rule), new Worker(string(s + 1)));
+    return true;
+}
+
 int
 main(int argc, char **argv)
 {
@@ -372,7 +373,8 @@ main(int argc, char **argv)
 	OPT_SAMPLE,
 	OPT_DATE_TERMS,
 	OPT_NO_DATE_TERMS,
-	OPT_READ_FILTERS
+	OPT_READ_FILTERS,
+	OPT_READ_WORKERS
     };
     constexpr auto NO_ARG = no_argument;
     constexpr auto REQ_ARG = required_argument;
@@ -382,13 +384,14 @@ main(int argc, char **argv)
 	{ "overwrite",		NO_ARG,		NULL, 'o' },
 	{ "duplicates",		REQ_ARG,	NULL, 'd' },
 	{ "no-delete",		NO_ARG,		NULL, 'p' },
-	{ "preserve-nonduplicates", NO_ARG,	NULL, 'p' },
 	{ "db",			REQ_ARG,	NULL, 'D' },
 	{ "url",		REQ_ARG,	NULL, 'U' },
 	{ "mime-type",		REQ_ARG,	NULL, 'M' },
 	{ "mime-type-match",	REQ_ARG,	NULL, 'G' },
 	{ "filter",		REQ_ARG,	NULL, 'F' },
+	{ "worker",		REQ_ARG,	NULL, 'W' },
 	{ "read-filters",	REQ_ARG,	NULL, OPT_READ_FILTERS },
+	{ "read-workers",	REQ_ARG,	NULL, OPT_READ_WORKERS },
 	{ "depth-limit",	REQ_ARG,	NULL, 'l' },
 	{ "follow",		NO_ARG,		NULL, 'f' },
 	{ "ignore-exclusions",	NO_ARG,		NULL, 'i' },
@@ -411,6 +414,7 @@ main(int argc, char **argv)
     map<string, string> mime_map;
 
     index_add_default_filters();
+    index_add_default_libraries();
 
     if (argc == 2 && strcmp(argv[1], "-v") == 0) {
 	// -v was the short option for --version in 1.2.3 and earlier, but
@@ -422,7 +426,8 @@ main(int argc, char **argv)
 
     string dbpath;
     int getopt_ret;
-    while ((getopt_ret = gnu_getopt_long(argc, argv, "hvd:D:U:M:G:F:l:s:pfRSVe:im:E:T:C",
+    while ((getopt_ret = gnu_getopt_long(argc, argv,
+					 "hvd:D:U:M:G:F:W:l:s:pfRSVe:im:E:T:C",
 					 longopts, NULL)) != -1) {
 	switch (getopt_ret) {
 	case 'h': {
@@ -437,8 +442,7 @@ main(int argc, char **argv)
 "  -d, --duplicates=ARG      set duplicate handling: ARG can be 'ignore' or\n"
 "                            'replace' (default: replace)\n"
 "  -p, --no-delete           skip the deletion of documents corresponding to\n"
-"                            deleted files (--preserve-nonduplicates is a\n"
-"                            deprecated alias for --no-delete)\n"
+"                            deleted files\n"
 "  -e, --empty-docs=ARG      how to handle documents we extract no text from:\n"
 "                            ARG can be index, warn (issue a diagnostic and\n"
 "                            index), or skip.  (default: warn)\n"
@@ -458,12 +462,23 @@ main(int argc, char **argv)
 "                            in a temporary file) with format T (Content-Type\n"
 "                            or file extension; currently txt (default), html\n"
 "                            or svg) in character encoding C (default: UTF-8).\n"
-"                            E.g. -Fapplication/octet-stream:'strings -n8'\n"
+"                            E.g. -Fapplication/octet-stream:'|strings -n8'\n"
 "                            or -Ftext/x-foo,,utf-16:'foo2utf16 %f %t'\n"
+"  -W, --worker=TYPE:WORKER  process files with MIME Content-Type TYPE using\n"
+"                            worker sub-process WORKER.  WORKER is the name of\n"
+"                            the program to run to start the worker. If it has\n"
+"                            no path then it's looked for in pkglibbindir (which\n"
+"                            can be overridden by setting environment variable\n"
+"                            XAPIAN_OMEGA_PKGLIBBINDIR).  This invocation will\n"
+"                            look in: " << get_pkglibbindir() << "\n"
 "      --read-filters=FILE   bulk-load --filter arguments from FILE, which\n"
 "                            should contain one such argument per line (e.g.\n"
 "                            text/x-bar:bar2txt --utf8).  Lines starting with #\n"
 "                            are treated as comments and ignored.\n"
+"      --read-workers=FILE   bulk-load --worker arguments from FILE, which\n"
+"                            should contain one such argument per line (e.g.\n"
+"                            text/x-bar:omindex_libbar).  Lines starting with #\n"
+"                            are treated as comments and ignored.\n"
 "  -l, --depth-limit=LIMIT   set recursion limit (0 = unlimited)\n"
 "  -f, --follow              follow symbolic links\n"
 "  -i, --ignore-exclusions   ignore meta robots tags and similar exclusions\n"
@@ -488,10 +503,10 @@ main(int argc, char **argv)
 "                            on Microsoft DFS shares.\n"
 "  -C, --track-ctime         track each file's ctime so we can detect changes\n"
 "                            to ownership or permissions.\n"
-"      --date-terms          ignored for forward compatibility with Omega 2.\n"
-"      --no-date-terms       don't index D, M and Y prefixed terms to support\n"
-"                            date range filtering using terms (we now recommend\n"
+"      --date-terms          index D, M and Y prefixed terms to support date\n"
+"                            range filtering using terms (we now recommend\n"
 "                            using a value slot for this instead).\n"
+"      --no-date-terms       ignored for compatibility with Omega 1.4.x.\n"
 "  -v, --verbose             show more information about what is happening\n"
 "      --overwrite           create the database anew (the default is to update\n"
 "                            if the database already exists)" << endl;
@@ -529,7 +544,10 @@ main(int argc, char **argv)
 	    delete_removed_documents = false;
 	    break;
 	case 'l': { // Set recursion limit
-	    int arg = atoi(optarg);
+	    int arg;
+	    if (!parse_signed(optarg, arg)) {
+		throw "Recursion limit must be an integer";
+	    }
 	    if (arg < 0) arg = 0;
 	    depth_limit = size_t(arg);
 	    break;
@@ -557,6 +575,10 @@ main(int argc, char **argv)
 	    if (!parse_filter_rule(optarg, mime_map))
 		return 1;
 	    break;
+	case 'W':
+	    if (!parse_worker_rule(optarg))
+		return 1;
+	    break;
 	case OPT_READ_FILTERS: {
 	    ifstream stream(optarg);
 	    if (!stream) {
@@ -575,6 +597,24 @@ main(int argc, char **argv)
 		return 1;
 	    break;
 	}
+	case OPT_READ_WORKERS: {
+	    ifstream stream(optarg);
+	    if (!stream) {
+		cerr << "Unable to open worker file '" << optarg << "' "
+			"(" << strerror(errno) << ')' << endl;
+		return 1;
+	    }
+	    string rule;
+	    bool all_ok = true;
+	    while (getline(stream, rule)) {
+		if (startswith(rule, '#')) continue;
+		if (!parse_worker_rule(rule.c_str()))
+		    all_ok = false;
+	    }
+	    if (!all_ok)
+		return 1;
+	    break;
+	}
 	case 'D':
 	    dbpath = optarg;
 	    break;
@@ -677,10 +717,10 @@ main(int argc, char **argv)
 	    use_ctime = true;
 	    break;
 	case OPT_DATE_TERMS:
-	    // Ignored for compatibility with Omega 2.0.0.
+	    date_terms = true;
 	    break;
 	case OPT_NO_DATE_TERMS:
-	    date_terms = false;
+	    // Ignored for compatibility with Omega 1.4.0.
 	    break;
 	case 'G': {
 	    char * s = strrchr(optarg, ':');
