@@ -5,7 +5,7 @@
  * Copyright 2001 James Aylett
  * Copyright 2001,2002 Ananova Ltd
  * Copyright 2002 Intercede 1749 Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2013,2014,2015,2016,2017,2018,2019,2020,2021 Olly Betts
+ * Copyright 2002-2026 Olly Betts
  * Copyright 2008 Thomas Viehmann
  *
  * This program is free software; you can redistribute it and/or
@@ -29,6 +29,7 @@
 #include <algorithm>
 #include <iostream>
 #include <map>
+#include <random>
 #include <set>
 #include <unordered_map>
 #include <unordered_set>
@@ -73,7 +74,8 @@
 #include "values.h"
 #include "weight.h"
 #include "expand.h"
-
+#include "md5wrap.h"
+#include "parseint.h"
 #include <xapian.h>
 
 using namespace std;
@@ -82,25 +84,6 @@ using Xapian::Utf8Iterator;
 
 using Xapian::Unicode::is_wordchar;
 
-#ifndef SNPRINTF
-#include <cstdarg>
-
-static int my_snprintf(char *str, size_t size, const char *format, ...)
-{
-    int res;
-    va_list ap;
-    va_start(ap, format);
-    str[size - 1] = '\0';
-    res = vsprintf(str, format, ap);
-    if (str[size - 1] || res < 0 || size_t(res) >= size)
-	abort(); /* Overflowed! */
-    va_end(ap);
-    return res;
-}
-#else
-#define my_snprintf SNPRINTF
-#endif
-
 /// Map shard to DB parameter value and stats to allow docid mapping.
 vector<SubDB> subdbs;
 
@@ -292,6 +275,12 @@ read_qp_flags(const string & opt_pfx, unsigned f)
 		    break;
 		}
 		break;
+	    case 'f':
+		if (strcmp(s, "fuzzy") == 0) {
+		    mask = Xapian::QueryParser::FLAG_FUZZY;
+		    break;
+		}
+		break;
 	    case 'l':
 		if (strcmp(s, "lovehate") == 0) {
 		    mask = Xapian::QueryParser::FLAG_LOVEHATE;
@@ -303,6 +292,10 @@ read_qp_flags(const string & opt_pfx, unsigned f)
 		    mask = Xapian::QueryParser::FLAG_NO_POSITIONS;
 		    break;
 		}
+		if (strcmp(s, "no_proper_noun_heuristic") == 0) {
+		    mask = Xapian::QueryParser::FLAG_NO_PROPER_NOUN_HEURISTIC;
+		    break;
+		}
 		if (strcmp(s, "ngrams") == 0) {
 		    mask = Xapian::QueryParser::FLAG_NGRAMS;
 		    break;
@@ -337,6 +330,24 @@ read_qp_flags(const string & opt_pfx, unsigned f)
 		    mask = Xapian::QueryParser::FLAG_WILDCARD;
 		    break;
 		}
+#if XAPIAN_AT_LEAST(1,5,0)
+		if (strcmp(s, "wildcard_glob") == 0) {
+		    mask = Xapian::QueryParser::FLAG_WILDCARD_GLOB;
+		    break;
+		}
+		if (strcmp(s, "wildcard_multi") == 0) {
+		    mask = Xapian::QueryParser::FLAG_WILDCARD_MULTI;
+		    break;
+		}
+		if (strcmp(s, "wildcard_single") == 0) {
+		    mask = Xapian::QueryParser::FLAG_WILDCARD_SINGLE;
+		    break;
+		}
+		if (strcmp(s, "word_breaks") == 0) {
+		    mask = Xapian::QueryParser::FLAG_WORD_BREAKS;
+		    break;
+		}
+#endif
 		break;
 	}
 
@@ -396,7 +407,7 @@ parse_queries(const string& oldp)
 	    // prefer the first user_prefix for which a particular term prefix
 	    // is specified.
 	    termprefix_to_userprefix.insert(make_pair(term_pfx, user_prefix));
-	} while (++i);
+	} while (UNSIGNED_OVERFLOW_OK(++i));
     }
     pfx = option.lower_bound("boolprefix,");
     for (; pfx != option.end() && startswith(pfx->first, "boolprefix,"); ++pfx) {
@@ -409,8 +420,6 @@ parse_queries(const string& oldp)
 
     try {
 	unsigned default_flags = read_qp_flags("flag_", 0);
-	if (option["spelling"] == "true")
-	    default_flags |= qp.FLAG_SPELLING_CORRECTION;
 
 	vector<Xapian::Query> queries;
 	queries.reserve(query_strings.size());
@@ -432,7 +441,22 @@ parse_queries(const string& oldp)
 	    if (!q.empty())
 		queries.push_back(q);
 	}
-	query = Xapian::Query(query.OP_AND, queries.begin(), queries.end());
+
+	Xapian::Query::op intra_query_op = Xapian::Query::OP_AND;
+	if (queries.size() > 1) {
+	    // Determine operator to use to combine multiple P and P.<prefix>
+	    // parameters.  Note that we only need to bother if there are two
+	    // or more query strings, since for one or none the operator
+	    // specified isn't actually used.
+	    opt_it = option.find("intra_query_op");
+	    if (opt_it != option.end()) {
+		const string& v = opt_it->second;
+		if (v == "OR" || v == "or") {
+		    intra_query_op = Xapian::Query::OP_OR;
+		}
+	    }
+	}
+	query = Xapian::Query(intra_query_op, queries.begin(), queries.end());
     } catch (Xapian::QueryParserError &e) {
 	error_msg = e.get_msg();
 	return BAD_QUERY;
@@ -965,6 +989,7 @@ CMD_foreach,
 CMD_freq,
 CMD_ge,
 CMD_gt,
+CMD_hash,
 CMD_highlight,
 CMD_hit,
 CMD_hitlist,
@@ -980,6 +1005,7 @@ CMD_json,
 CMD_jsonarray,
 CMD_jsonbool,
 CMD_jsonobject,
+CMD_jsonobject2,
 CMD_keys,
 CMD_last,
 CMD_lastpage,
@@ -1015,6 +1041,7 @@ CMD_prettyurl,
 CMD_query,
 CMD_querydescription,
 CMD_queryterms,
+CMD_random,
 CMD_range,
 CMD_record,
 CMD_relevant,
@@ -1029,6 +1056,7 @@ CMD_snippet,
 CMD_sort,
 CMD_sortableunserialise,
 CMD_split,
+CMD_srandom,
 CMD_stoplist,
 CMD_sub,
 CMD_subdb,
@@ -1105,7 +1133,7 @@ T(eq,		   2, 2, N, 0), // test equality
 T(error,	   0, 0, N, 0), // error message
 T(field,	   1, 2, N, 0), // lookup field in record
 T(filesize,	   1, 1, N, 0), // pretty printed filesize
-T(filters,	   0, 0, N, 0), // serialisation of current filters
+T(filters,	   0, 1, N, 0), // serialisation of current filters
 T(filterterms,	   1, 1, N, 0), // list of terms with a given prefix
 T(find,		   2, 2, N, 0), // find entry in list
 T(fmt,		   0, 0, N, 0), // name of current format
@@ -1113,6 +1141,7 @@ T(foreach,	   2, 2, 1, 0), // evaluate something for every entry in a list
 T(freq,		   1, 1, N, 0), // frequency of a term
 T(ge,		   2, 2, N, 0), // test >=
 T(gt,		   2, 2, N, 0), // test >
+T(hash,		   2, 2, N, 0), // hash a string using the specified hash function
 T(highlight,	   2, 4, N, 0), // html escape and highlight words from list
 T(hit,		   0, 0, N, 0), // hit number of current mset entry (0-based)
 T(hitlist,	   1, 1, 0, M), // display hitlist using format in argument
@@ -1128,6 +1157,7 @@ T(json,		   1, 1, N, 0), // JSON string escaping
 T(jsonarray,	   1, 2, 1, 0), // Format list as a JSON array
 T(jsonbool,	   1, 1, 1, 0), // Format list as a JSON bool
 T(jsonobject,	   1, 3, 1, 0), // Format map as JSON object
+T(jsonobject2,	   2, 4, 2, 0), // Format 2 lists as JSON object
 T(keys,		   1, 1, N, 0), // list of keys from a map
 T(last,		   0, 0, N, M), // hit number one beyond end of current page
 T(lastpage,	   0, 0, N, M), // number of last hit page
@@ -1163,6 +1193,7 @@ T(prettyurl,	   1, 1, N, 0), // pretty version of URL
 T(query,	   0, 1, N, Q), // query
 T(querydescription,0, 0, N, M), // query.get_description() (run_query() adds filters so M)
 T(queryterms,	   0, 0, N, Q), // list of query terms
+T(random,	   1, 1, N, 0), // return a random number
 T(range,	   2, 2, N, 0), // return list of values between start and end
 T(record,	   0, 1, N, 0), // record contents of document
 T(relevant,	   0, 1, N, Q), // is document relevant?
@@ -1173,11 +1204,12 @@ T(seterror,	   1, 1, N, 0), // set error_msg, setting it early stops query execu
 T(setmap,	   1, N, N, 0), // set map of option values
 T(setrelevant,	   1, 1, N, Q), // set rset
 T(slice,	   2, 2, N, 0), // slice a list using a second list
-T(snippet,	   1, 2, N, M), // generate snippet from text
+T(snippet,	   1, 6, N, M), // generate snippet from text
 T(sort,		   1, 2, N, 0), // alpha sort a list
 T(sortableunserialise,
 		   1, 1, N, 0), // decode with Xapian::sortable_unserialise
 T(split,	   1, 2, N, 0), // split a string to give a list
+T(srandom,	   1, 1, N, 0), // seed for random number
 T(stoplist,	   0, 0, N, Q), // return list of stopped terms
 T(sub,		   2, 2, N, 0), // subtract
 T(subdb,	   0, 1, N, 0), // name of subdb docid is in
@@ -1232,6 +1264,10 @@ write_all(int fd, const char * buf, size_t count)
     return 0;
 }
 
+// mersenne twister for RNG
+static mt19937 rng;
+static bool seed_set = false;
+
 static string eval(const string& fmt, vector<string>& param);
 
 /** Implements $foreach{} and $map{}. */
@@ -1396,7 +1432,10 @@ eval(const string& fmt, vector<string>& param)
 	    case CMD_allterms: {
 		// list of all terms indexing document
 		Xapian::docid id = q0;
-		if (!args.empty()) id = string_to_int(args[0]);
+		if (!args.empty() &&
+		    (!parse_unsigned(args[0].c_str(), id) || id == 0)) {
+		    throw "Document id for command allterms should be > 0";
+		}
 		for (Xapian::TermIterator term = db.termlist_begin(id);
 		     term != db.termlist_end(id); ++term) {
 		    value += *term;
@@ -1424,7 +1463,7 @@ eval(const string& fmt, vector<string>& param)
 		value.reserve((input.size() + 2) / 3 * 4);
 		auto it = input.begin();
 		auto n = input.size() / 3;
-		while (n--) {
+		for ( ; n; --n) {
 		    uint32_t v = uint8_t(*it++);
 		    v = (v << 8) | uint8_t(*it++);
 		    v = (v << 8) | uint8_t(*it++);
@@ -1479,9 +1518,14 @@ eval(const string& fmt, vector<string>& param)
 		if (!value.empty()) value.erase(value.size() - 1);
 		break;
 	    }
-	    case CMD_chr:
-		Xapian::Unicode::append_utf8(value, string_to_int(args[0]));
+	    case CMD_chr: {
+		unsigned int codepoint;
+		if (!parse_unsigned(args[0].c_str(), codepoint)) {
+		    throw "Unicode codepoint for command chr should be >= 0";
+		}
+		Xapian::Unicode::append_utf8(value, codepoint);
 		break;
+	    }
 	    case CMD_collapsed: {
 		value = str(collapsed);
 		break;
@@ -1518,7 +1562,11 @@ eval(const string& fmt, vector<string>& param)
 		value = args[0];
 		if (!value.empty()) {
 		    char buf[64] = "";
-		    time_t date = string_to_int(value);
+		    time_t date;
+		    if (!parse_signed(value.c_str(), date)) {
+			throw "Date (in secs) for command date should "
+			      "be an integer";
+		    }
 		    if (date != static_cast<time_t>(-1)) {
 			struct tm *then;
 			then = gmtime(&date);
@@ -1594,13 +1642,20 @@ eval(const string& fmt, vector<string>& param)
 		break;
 	    case CMD_field: {
 		Xapian::docid did = q0;
-		if (args.size() > 1) did = string_to_int(args[1]);
+		if (args.size() > 1 &&
+		    (!parse_unsigned(args[1].c_str(), did) || did == 0)) {
+		    throw "Document id for command field should be > 0";
+		}
 		value = fields.get_field(did, args[0]);
 		break;
 	    }
 	    case CMD_filesize: {
+		if (args[0].empty()) break;
 		// FIXME: rounding?  i18n?
-		int size = string_to_int(args[0]);
+		int size;
+		if (!parse_signed(args[0].c_str(), size)) {
+		    throw "Filesize must be an integer";
+		}
 		int intpart = size;
 		int fraction = -1;
 		const char * format = 0;
@@ -1629,10 +1684,10 @@ eval(const string& fmt, vector<string>& param)
 		    char buf[200];
 		    int len;
 		    if (fraction == -1) {
-			len = my_snprintf(buf, sizeof(buf), format, intpart);
+			len = snprintf(buf, sizeof(buf), format, intpart);
 		    } else {
 			fraction = (fraction * 10 / 1024) + '0';
-			len = my_snprintf(buf, sizeof(buf), format, intpart, fraction);
+			len = snprintf(buf, sizeof(buf), format, intpart, fraction);
 		    }
 		    if (len < 0 || unsigned(len) > sizeof(buf)) len = sizeof(buf);
 		    value.assign(buf, len);
@@ -1640,7 +1695,7 @@ eval(const string& fmt, vector<string>& param)
 		break;
 	    }
 	    case CMD_filters:
-		value = filters;
+		value = args.size() ? old_filters : filters;
 		break;
 	    case CMD_filterterms: {
 		Xapian::TermIterator term = db.allterms_begin();
@@ -1686,14 +1741,7 @@ eval(const string& fmt, vector<string>& param)
 		const string& term = args[0];
 		Xapian::doccount termfreq = 0;
 		if (done_query) {
-		    try {
-			termfreq = mset.get_termfreq(term);
-		    } catch (const Xapian::InvalidOperationError&) {
-			// In 1.4.x and earlier, InvalidOperationError is
-			// thrown if the MSet is empty and not associated with
-			// an Enquire object.  In 2.0.0 and later, a termfreq
-			// of 0 is returned for this case.
-		    }
+		    termfreq = mset.get_termfreq(term);
 		}
 		if (termfreq == 0) {
 		    // We want $freq to work before the match is run, and we
@@ -1711,6 +1759,22 @@ eval(const string& fmt, vector<string>& param)
 		if (string_to_int(args[0]) > string_to_int(args[1]))
 		    value = "true";
 		break;
+	    case CMD_hash: {
+		const string& data = args[0];
+		const string& hash = args[1];
+		if (hash == "md5") {
+		    string md5;
+		    md5_string(data, md5);
+		    value.reserve(md5.size() * 2);
+		    for (unsigned char byte : md5) {
+			value += "0123456789abcdef"[byte >> 4];
+			value += "0123456789abcdef"[byte & 0x0f];
+		    }
+		} else {
+		    throw "Unknown hash function: " + hash;
+		}
+		break;
+	    }
 	    case CMD_highlight: {
 		string bra, ket;
 		if (args.size() > 2) {
@@ -1871,8 +1935,7 @@ eval(const string& fmt, vector<string>& param)
 		value = args[0].empty() ? "false" : "true";
 		break;
 	    case CMD_jsonobject: {
-		vector<string> new_args;
-		new_args.push_back(string());
+		vector<string> new_args(1);
 
 		class map_range {
 		    typedef map<string, string>::const_iterator iterator;
@@ -1912,6 +1975,109 @@ eval(const string& fmt, vector<string>& param)
 				});
 		break;
 	    }
+	    case CMD_jsonobject2: {
+		vector<string> new_args(1);
+
+		static string dummy;
+
+		class list_range {
+		    const string& keys;
+		    const string& values;
+
+		  public:
+		    class iterator {
+			const string& keys;
+			const string& values;
+			string::size_type ki = 0;
+			string::size_type kj;
+			string::size_type vi = 0;
+			string::size_type vj;
+
+		      public:
+			iterator()
+			    : keys(dummy), values(dummy),
+			      ki(string::npos), vi(string::npos) {}
+
+			iterator(const string& k, const string& v)
+			    : keys(k), values(v) {
+			    if (keys.empty() && values.empty()) {
+				// Don't treat this as: { "": "" }
+				ki = kj = vi = vj = string::npos;
+			    } else {
+				kj = keys.find('\t');
+				vj = values.find('\t');
+			    }
+			}
+
+			pair<string, string> operator*() const {
+			    return {keys.substr(ki, kj - ki),
+				    values.substr(vi, vj - vi)};
+			}
+
+			iterator& operator++() {
+			    ki = kj;
+			    if (ki != string::npos) {
+				++ki;
+				kj = keys.find('\t', ki);
+			    }
+			    vi = vj;
+			    if (vi != string::npos) {
+				++vi;
+				vj = values.find('\t', vi);
+			    }
+			    if ((ki == string::npos) !=
+				(vi == string::npos)) {
+				throw "$jsonobject2: Different number of keys "
+				      "and values";
+			    }
+			    return *this;
+			}
+
+			iterator operator++(int) {
+			    iterator r = *this;
+			    operator++();
+			    return r;
+			}
+
+			bool operator==(const iterator& o) const {
+			    return ki == o.ki && vi == o.vi;
+			}
+
+			bool operator!=(const iterator& o) const {
+			    return !(*this == o);
+			}
+		    };
+
+		    list_range(const string& k, const string& v)
+			: keys(k), values(v) { }
+
+		    iterator begin() const { return iterator(keys, values); }
+		    iterator end() const { return iterator(); }
+		};
+
+		value = to_json(list_range(args[0], args[1]),
+				[&](const string& k) {
+				    string key = k;
+				    if (args.size() > 2 && !args[2].empty()) {
+					new_args[0] = std::move(key);
+					key = eval(args[2], new_args);
+				    }
+				    return key;
+				},
+				[&](const string& v) {
+				    if (args.size() > 3 && !args[3].empty()) {
+					new_args[0] = v;
+					return eval(args[3], new_args);
+				    }
+				    string r(1, '"');
+				    string elt = v;
+				    json_escape(elt);
+				    r += elt;
+				    r += '"';
+				    return r;
+				});
+		break;
+	    }
 	    case CMD_keys: {
 		string prefix = args[0] + ',';
 		auto i = option.lower_bound(prefix);
@@ -1981,10 +2147,17 @@ eval(const string& fmt, vector<string>& param)
 		break;
 	    }
 	    case CMD_log: {
-		if (!vet_filename(args[0])) break;
+		if (!vet_filename(args[0])) {
+		    value = "filename can't contain \"..\"";
+		    break;
+		}
 		string logfile = log_dir + args[0];
 		int fd = open(logfile.c_str(), O_CREAT|O_APPEND|O_WRONLY, 0644);
-		if (fd == -1) break;
+		if (fd == -1) {
+		    value = "open failed: ";
+		    value += strerror(errno);
+		    break;
+		}
 		vector<string> noargs;
 		noargs.resize(1);
 		string line;
@@ -1995,7 +2168,10 @@ eval(const string& fmt, vector<string>& param)
 		}
 		line = eval(line, noargs);
 		line += '\n';
-		(void)write_all(fd, line.data(), line.length());
+		if (write_all(fd, line.data(), line.length()) < 0) {
+		    value = "write failed: ";
+		    value += strerror(errno);
+		}
 		close(fd);
 		break;
 	    }
@@ -2142,9 +2318,15 @@ eval(const string& fmt, vector<string>& param)
 		}
 		break;
 	    }
-	    case CMD_pack:
-		value = int_to_binary_string(string_to_int(args[0]));
+	    case CMD_pack: {
+		int number;
+		if (!parse_signed(args[0].c_str(), number)) {
+		    throw "NUMBER parameter for pack command "
+			  "must be an integer";
+		}
+		value = int_to_binary_string(number);
 		break;
+	    }
 	    case CMD_percentage:
 		// percentage score
 		value = str(percent);
@@ -2178,9 +2360,27 @@ eval(const string& fmt, vector<string>& param)
 	    case CMD_queryterms:
 		value = queryterms;
 		break;
+	    case CMD_random: {
+		if (!seed_set) {
+		    random_device rd;
+		    rng.seed(rd());
+		    seed_set = true;
+		}
+		uniform_int_distribution<int>
+		    distr(0, string_to_int(args[0]));
+		value = str(distr(rng));
+		break;
+	    }
 	    case CMD_range: {
-		int start = string_to_int(args[0]);
-		int end = string_to_int(args[1]);
+		int start, end;
+		if (!parse_signed(args[0].c_str(), start)) {
+		    throw "Start value for range command "
+			  "must be an integer";
+		}
+		if (!parse_signed(args[1].c_str(), end)) {
+		    throw "End value for range command "
+			  "must be an integer";
+		}
 		while (start <= end) {
 		    value += str(start);
 		    if (start < end) value += '\t';
@@ -2190,14 +2390,20 @@ eval(const string& fmt, vector<string>& param)
 	    }
 	    case CMD_record: {
 		Xapian::docid id = q0;
-		if (!args.empty()) id = string_to_int(args[0]);
+		if (!args.empty() &&
+		    (!parse_unsigned(args[0].c_str(), id) || id == 0)) {
+		    throw "Document id for command record should be > 0";
+		}
 		value = db.get_document(id).get_data();
 		break;
 	    }
 	    case CMD_relevant: {
 		// document id if relevant; empty otherwise
 		Xapian::docid id = q0;
-		if (!args.empty()) id = string_to_int(args[0]);
+		if (!args.empty() &&
+		    (!parse_unsigned(args[0].c_str(), id) || id == 0)) {
+		    throw "Document id for command relevant should be > 0";
+		}
 		auto i = ticked.find(id);
 		if (i != ticked.end()) {
 		    i->second = false; // icky side-effect
@@ -2275,15 +2481,65 @@ eval(const string& fmt, vector<string>& param)
 	    }
 	    case CMD_snippet: {
 		size_t length = 200;
-		if (args.size() > 1) {
-		    length = string_to_int(args[1]);
+		if (args.size() > 1 && !args[1].empty()) {
+		    if (!parse_unsigned(args[1].c_str(), length)) {
+			throw "Snippet length must be >= 0";
+		    }
+		}
+		unsigned flags = mset.SNIPPET_BACKGROUND_MODEL |
+				 mset.SNIPPET_EXHAUSTIVE;
+		if (args.size() > 2 && !args[2].empty()) {
+		    flags = 0;
+		    const string& s = args[2];
+		    size_t i = 0;
+		    while (true) {
+			size_t j = s.find('|', i);
+			string flag(s, i, j - i);
+			for (char& c : flag) {
+			    c = C_tolower(c);
+			}
+			if (startswith(flag, "snippet_")) {
+			    flag.erase(0, CONST_STRLEN("snippet_"));
+			}
+			if (flag == "background_model") {
+			    flags |= mset.SNIPPET_BACKGROUND_MODEL;
+			} else if (flag == "cjk_ngram") {
+			    flags |= mset.SNIPPET_CJK_NGRAM;
+			} else if (flag == "empty_without_match") {
+			    flags |= mset.SNIPPET_EMPTY_WITHOUT_MATCH;
+			} else if (flag == "exhaustive") {
+			    flags |= mset.SNIPPET_EXHAUSTIVE;
+			} else if (flag == "ngrams") {
+			    flags |= mset.SNIPPET_NGRAMS;
+			} else if (flag == "word_breaks") {
+			    flags |= mset.SNIPPET_WORD_BREAKS;
+			} else {
+			    throw "Unknown $snippet flag '" + flag + "'";
+			}
+			if (j == string::npos) break;
+			i = j + 1;
+		    }
+		}
+		string bra, ket, gap;
+		if (args.size() > 3) {
+		    bra = args[3];
+		} else {
+		    bra = "<strong>";
+		}
+		if (args.size() > 4) {
+		    ket = args[4];
+		} else {
+		    ket = "</strong>";
+		}
+		if (args.size() > 5) {
+		    gap = args[5];
+		} else {
+		    gap = "...";
 		}
 		if (!stemmer)
 		    stemmer = new Xapian::Stem(option["stemmer"]);
-		// FIXME: Allow start and end highlight and omit to be specified.
-		value = mset.snippet(args[0], length, *stemmer,
-				     mset.SNIPPET_BACKGROUND_MODEL|mset.SNIPPET_EXHAUSTIVE,
-				     "<strong>", "</strong>", "...");
+		value = mset.snippet(args[0], length, *stemmer, flags,
+				     bra, ket, gap);
 		break;
 	    }
 	    case CMD_sort:
@@ -2317,6 +2573,12 @@ eval(const string& fmt, vector<string>& param)
 		}
 		break;
 	    }
+	    case CMD_srandom: {
+		int seed = string_to_int(args[0]);
+		rng.seed(seed);
+		seed_set = true;
+		break;
+	    }
 	    case CMD_stoplist: {
 		Xapian::TermIterator i = qp.stoplist_begin();
 		Xapian::TermIterator end = qp.stoplist_end();
@@ -2332,13 +2594,19 @@ eval(const string& fmt, vector<string>& param)
 		break;
 	    case CMD_subdb: {
 		Xapian::docid id = q0;
-		if (args.size() > 0) id = string_to_int(args[0]);
+		if (args.size() > 0 &&
+		    (!parse_unsigned(args[0].c_str(), id) || id == 0)) {
+		    throw "Document id of the subdb command should be > 0";
+		}
 		value = subdbs[(id - 1) % subdbs.size()].get_name();
 		break;
 	    }
 	    case CMD_subid: {
 		Xapian::docid id = q0;
-		if (args.size() > 0) id = string_to_int(args[0]);
+		if (args.size() > 0 &&
+		    (!parse_unsigned(args[0].c_str(), id) || id == 0)) {
+		    throw "Document id of the subid command should be > 0";
+		}
 		// This is the docid in the single shard.
 		Xapian::docid shard_did = (id - 1) / subdbs.size() + 1;
 		// We now need to map this back to the docid in the collection
@@ -2349,7 +2617,11 @@ eval(const string& fmt, vector<string>& param)
 		break;
 	    }
 	    case CMD_substr: {
-		int start = string_to_int(args[1]);
+		int start;
+		if (!parse_signed(args[1].c_str(), start)) {
+		    throw "Start value for substr command "
+			  "must be an integer";
+		}
 		if (start < 0) {
 		    if (static_cast<size_t>(-start) >= args[0].size()) {
 			start = 0;
@@ -2361,7 +2633,11 @@ eval(const string& fmt, vector<string>& param)
 		}
 		size_t len = string::npos;
 		if (args.size() > 2) {
-		    int int_len = string_to_int(args[2]);
+		    int int_len;
+		    if (!parse_signed(args[2].c_str(), int_len)) {
+			throw "Length value for substr command "
+			      "must be an integer";
+		    }
 		    if (int_len >= 0) {
 			len = size_t(int_len);
 		    } else {
@@ -2435,7 +2711,7 @@ eval(const string& fmt, vector<string>& param)
 	    case CMD_time:
 		if (secs >= 0) {
 		    char buf[64];
-		    my_snprintf(buf, sizeof(buf), "%.6f", secs);
+		    snprintf(buf, sizeof(buf), "%.6f", secs);
 		    // MSVC's snprintf omits the zero byte if the string is
 		    // sizeof(buf) long.
 		    buf[sizeof(buf) - 1] = '\0';
@@ -2449,9 +2725,13 @@ eval(const string& fmt, vector<string>& param)
 	    case CMD_topterms:
 		if (enquire) {
 		    int howmany = 16;
-		    if (!args.empty()) howmany = string_to_int(args[0]);
+		    if (!args.empty()) {
+			if (!parse_signed(args[0].c_str(), howmany)) {
+			    throw "Number of terms for command "
+				  "topterms must be an integer";
+			}
+		    }
 		    if (howmany < 0) howmany = 0;
-
 		    // List of expand terms
 		    Xapian::ESet eset;
 		    OmegaExpandDecider decider(db, &termset);
@@ -2492,12 +2772,17 @@ eval(const string& fmt, vector<string>& param)
 	    case CMD_transform:
 		omegascript_transform(value, args);
 		break;
-	    case CMD_truncate:
+	    case CMD_truncate: {
+		unsigned int length;
+		if (!parse_unsigned(args[1].c_str(), length)) {
+		    throw "Length for truncate command must be >= 0";
+		}
 		value = generate_sample(args[0],
-					string_to_int(args[1]),
+					length,
 					args.size() > 2 ? args[2] : string(),
 					args.size() > 3 ? args[3] : string());
 		break;
+	    }
 	    case CMD_uniq: {
 		const string &list = args[0];
 		if (list.empty()) break;
@@ -2513,7 +2798,7 @@ eval(const string& fmt, vector<string>& param)
 			value += item;
 		    }
 		    prev = item;
-		    split = split2 + 1;
+		    split = UNSIGNED_OVERFLOW_OK(split2 + 1);
 		} while (split2 != string::npos);
 		break;
 	    }
@@ -2530,7 +2815,7 @@ eval(const string& fmt, vector<string>& param)
 			    value += '\t';
 			value += item;
 		    }
-		    split = split2 + 1;
+		    split = UNSIGNED_OVERFLOW_OK(split2 + 1);
 		} while (split2 != string::npos);
 		break;
 	    }
@@ -2561,18 +2846,30 @@ eval(const string& fmt, vector<string>& param)
 		break;
 	    case CMD_value: {
 		Xapian::docid id = q0;
-		Xapian::valueno value_no = string_to_int(args[0]);
-		if (args.size() > 1) id = string_to_int(args[1]);
-		value = db.get_document(id).get_value(value_no);
+		Xapian::valueno slot;
+		if (!parse_unsigned(args[0].c_str(), slot)) {
+		    throw "Value slot number should be >= 0";
+		}
+		if (args.size() > 1 &&
+		    (!parse_unsigned(args[1].c_str(), id) || id == 0)) {
+		    throw "Document id for value command must be > 0";
+		}
+		value = db.get_document(id).get_value(slot);
 		break;
 	    }
 	    case CMD_valuelowerbound: {
-		Xapian::valueno slot = string_to_int(args[0]);
+		Xapian::valueno slot;
+		if (!parse_unsigned(args[0].c_str(), slot)) {
+		    throw "Value slot number should be >= 0";
+		}
 		value = db.get_value_lower_bound(slot);
 		break;
 	    }
 	    case CMD_valueupperbound: {
-		Xapian::valueno slot = string_to_int(args[0]);
+		Xapian::valueno slot;
+		if (!parse_unsigned(args[0].c_str(), slot)) {
+		    throw "Value slot number should be >= 0";
+		}
 		value = db.get_value_upper_bound(slot);
 		break;
 	    }
@@ -2777,7 +3074,9 @@ ensure_query_parsed()
 	// to display
 	val = cgi_params.find("TOPDOC");
 	if (val != cgi_params.end()) {
-	    topdoc = atol(val->second.c_str());
+	    if (!parse_unsigned(val->second.c_str(), topdoc)) {
+		throw "TOPDOC parameter must be >= 0";
+	    }
 	}
 
 	// Handle next, previous, and page links
@@ -2790,6 +3089,9 @@ ensure_query_parsed()
 		topdoc = 0;
 	} else if ((val = cgi_params.find("[")) != cgi_params.end() ||
 		   (val = cgi_params.find("#")) != cgi_params.end()) {
+	    if (!C_isdigit(val->second[0])) {
+		throw "Page parameter must be >= 0";
+	    }
 	    long page = atol(val->second.c_str());
 	    // Do something sensible for page 0 (we count pages from 1).
 	    if (page == 0) page = 1;
@@ -2804,7 +3106,11 @@ ensure_query_parsed()
 	bool raw_search = false;
 	val = cgi_params.find("RAWSEARCH");
 	if (val != cgi_params.end()) {
-	    raw_search = bool(atol(val->second.c_str()));
+	    unsigned int temp;
+	    if (!parse_unsigned(val->second.c_str(), temp)) {
+		throw "RAWSEARCH parameter must be >= 0";
+	    }
+	    raw_search = bool(temp);
 	}
 
 	if (!raw_search) topdoc = (topdoc / hits_per_page) * hits_per_page;
@@ -2817,7 +3123,10 @@ ensure_query_parsed()
 	    const string & value = i->second;
 	    for (size_t j = 0; j < value.size(); j = value.find('.', j)) {
 		while (value[j] == '.') ++j;
-		Xapian::docid d = atoi(value.c_str() + j);
+		Xapian::docid d;
+		if (!parse_unsigned(value.c_str() + j, d) || d == 0) {
+		    throw "Document id for 'R' parameter must be > 0";
+		}
 		if (d) {
 		    rset.add_document(d);
 		    ticked[d] = true;
