@@ -73,11 +73,11 @@ std::string pretty_term(std::string term);
 
 class OmegaExpandDecider : public Xapian::ExpandDecider {
     Xapian::Database db;
-    set<string> exclude_stems;
+    std::set<std::string> exclude_stems;
   public:
-    OmegaExpandDecider(const Xapian::Database & db_,
-		       set<string> * querytermset = NULL);
-    bool operator()(const string& term) const override;
+    OmegaExpandDecider(const Xapian::Database& db_,
+		       std::set<std::string>* querytermset = NULL);
+    bool operator()(const std::string& term) const override;
 };
 
 std::string html_escape(const std::string &str);
