@@ -43,9 +43,11 @@
 #include <ctime>
 
 #include "commonhelp.h"
+#include "datetime.h"
+#include "genericxmlparser.h"
 #include "hashterm.h"
+#include "htmlparser.h"
 #include "loadfile.h"
-#include "myhtmlparse.h"
 #include "parseint.h"
 #include "setenv.h"
 #include "str.h"
@@ -53,7 +55,6 @@
 #include "strptime.h"
 #include "timegm.h"
 #include "utf8truncate.h"
-#include "utils.h"
 #include "values.h"
 
 #include "gnu_getopt.h"
@@ -77,7 +78,7 @@ static enum {
     UNIQUE_NEW,
     UNIQUE_WARN_SKIP,
     UNIQUE_SKIP
-} unique_missing = UNIQUE_WARN_NEW;
+} unique_missing = UNIQUE_ERROR;
 
 /// Track if UNIQUE action is unused in the current record.
 static bool unique_unused;
@@ -118,6 +119,7 @@ const char * action_names[] = {
     "truncate",
     "unhtml",
     "unique",
+    "unxml",
     "value",
     "valuenumeric",
     "valuepacked",
@@ -154,6 +156,7 @@ class Action {
 	TRUNCATE,
 	UNHTML,
 	UNIQUE,
+	UNXML,
 	VALUE,
 	VALUENUMERIC,
 	VALUEPACKED,
@@ -162,13 +165,13 @@ class Action {
     enum { SPLIT_NONE, SPLIT_DEDUP, SPLIT_SORT, SPLIT_PREFIXES };
   private:
     type action;
-    int num_arg;
+    int num_arg = 0;
     string string_arg;
     // Offset into indexscript line.
     size_t pos;
   public:
     Action(type action_, size_t pos_)
-	: action(action_), num_arg(0), pos(pos_) { }
+	: action(action_), pos(pos_) { }
     Action(type action_, size_t pos_, const string & arg)
 	: action(action_), string_arg(arg), pos(pos_) {
 	num_arg = atoi(string_arg.c_str());
@@ -303,7 +306,7 @@ getline_portable(istream& stream, string& line)
     istream& result = getline(stream, line);
     // Trim multiple \r characters, since that seems the best way to handle
     // that case.
-    line.resize(line.find_last_not_of('\r') + 1);
+    line.resize(UNSIGNED_OVERFLOW_OK(line.find_last_not_of('\r') + 1));
     return result;
 }
 
@@ -463,6 +466,8 @@ parse_index_script(const string &filename)
 			    code = Action::UNIQUE;
 			    min_args = 1;
 			    max_args = 2;
+			} else if (action == "unxml") {
+			    code = Action::UNXML;
 			}
 			break;
 		    case 'v':
@@ -913,6 +918,7 @@ bad_hex_digit:
 		case Action::TRIM:
 		case Action::TRUNCATE:
 		case Action::UNHTML:
+		case Action::UNXML:
 		    done = false;
 		    report_useless_action(filename, line_no,
 					  actions.back().get_pos(),
@@ -991,8 +997,6 @@ run_actions(vector<Action>::const_iterator action_it,
 		abort();
 	    case Action::NEW:
 		value = old_value;
-		// We're processing the same field again - give it a reprieve.
-		this_field_is_content = true;
 		break;
 	    case Action::FIELD:
 		if (!value.empty()) {
@@ -1225,14 +1229,14 @@ run_actions(vector<Action>::const_iterator action_it,
 		break;
 	    }
 	    case Action::UNHTML: {
-		MyHtmlParser p;
+		HtmlParser p;
 		try {
 		    // Default HTML character set is latin 1, though
 		    // not specifying one is deprecated these days.
-		    p.parse_html(value, "iso-8859-1", false);
+		    p.parse(value, "iso-8859-1", false);
 		} catch (const string & newcharset) {
 		    p.reset();
-		    p.parse_html(value, newcharset, true);
+		    p.parse(value, newcharset, true);
 		}
 		if (p.indexing_allowed)
 		    value = p.dump;
@@ -1240,6 +1244,12 @@ run_actions(vector<Action>::const_iterator action_it,
 		    value = "";
 		break;
 	    }
+	    case Action::UNXML: {
+		GenericXmlParser p;
+		p.parse(value);
+		value = std::move(p.dump);
+		break;
+	    }
 	    case Action::UNIQUE: {
 		unique_unused = false;
 
@@ -1545,6 +1555,7 @@ index_file(const char *fname, istream &stream,
     database.commit();
 }
 
+[[noreturn]]
 static void
 show_help(int exit_code)
 {
