@@ -23,9 +23,6 @@
 
 #include <string>
 
-/** Converts year, month, day into an 8 character string like: "20061031". */
-std::string date_to_string(int year, int month, int day);
-
 /** Converts a double to a string. */
 std::string double_to_string(double value);
 
